import os, sys
from random import randint, choice
from math import sin, cos, radians

import pygame
import Queue
from pygame.sprite import Sprite
from pygame.locals import *

from vec2d import vec2d

# Game Parameters
SCREEN_DIMENSION = 710, 500
BGROUND_FILENAME = 'bground.jpg'
ANT_FILENAMES = ['ant1.png', 'ant2.png', 'ant1.png', 'ant4.png']
GRIDSTART_X, GRIDSTART_Y = 25, 25
GRIDCOLS, GRIDROWS = 33, 21
GRIDSIZE = 20


def gridToScreen(gridCoord):
    return (GRIDSTART_X+gridCoord[0]*GRIDSIZE, GRIDSTART_Y+gridCoord[1]*GRIDSIZE)
def screenToGrid(screenCord):
    return ( (screenCord[0]-GRIDSTART_X)/GRIDSIZE, (screenCord[1]-GRIDSTART_Y)/GRIDSIZE )

class Attacker(Sprite):
    def __init__(self, screen, img_filenames, pathMatrix, speed, health):
	Sprite.__init__(self)
        self.screen = screen
        self.speed = speed
	
	self.base_images = []
	for filename in img_filenames:
	    self.base_images.append(pygame.image.load(filename).convert_alpha())
	    
	self.animFrame = 0
	self.image = self.base_images[self.animFrame]
	
        self.pos = vec2d(GRIDSTART_X+GRIDSIZE/2,GRIDSTART_Y+GRIDSIZE/2)
	self.direction = vec2d(1,1).normalized()
	self.pathMatrix = pathMatrix.deepCopy()
	self.pathMatrix.setByXY(0,0,0)
	self.waypoint = vec2d(0,0)
	self.navigate()

    def navigate(self):
	gridPos = screenToGrid( (self.pos.x, self.pos.y) )
	# check for 1s in adjacent cells
	if gridPos[1]>0 and self.pathMatrix.getByXY(gridPos[0], gridPos[1]-1)==1:    # if up node exists and there is a 1 there
	    navX = 0
	    navY = -1
	elif gridPos[0]<GRIDCOLS-1 and self.pathMatrix.getByXY(gridPos[0]+1, gridPos[1])==1:    # if right node exists and there is a 1 there
	    navX = 1
	    navY = 0
	elif gridPos[1]<GRIDROWS-1 and self.pathMatrix.getByXY(gridPos[0], gridPos[1]+1)==1:    # if down node exists and there is a 1 there
	    navX = 0
	    navY = 1
	elif gridPos[0]>0 and self.pathMatrix.getByXY(gridPos[0]-1, gridPos[1])==1:    # if left node exists and there is a 1 there
	    navX = -1
	    navY = 0
	# # Diagonals # #
	elif gridPos[0]<GRIDCOLS-1 and gridPos[1]<GRIDROWS-1 and self.pathMatrix.getByXY(gridPos[0]+1, gridPos[1]+1)==1:    #down/right
	    navX = 1
	    navY = 1
	elif gridPos[1]<GRIDROWS-1 and gridPos[0]>0 and self.pathMatrix.getByXY(gridPos[0]-1, gridPos[1]+1)==1:    #down/left
	    navX = -1
	    navY = 1
	elif gridPos[1]>0 and gridPos[0]>0 and gridPos[0]>0 and self.pathMatrix.getByXY(gridPos[0]-1, gridPos[1]-1)==1:    #up/left
	    navX = -1
	    navY = -1
	elif gridPos[1]>0 and gridPos[0]<GRIDCOLS-1 and self.pathMatrix.getByXY(gridPos[0]+1, gridPos[1]-1)==1:    #up/right
	    navX = 1
	    navY = -1
	else:
	    return True

	self.direction = vec2d(navX, navY).normalized()
	
	# determine and set waypoint for next turn
	wpGrid = vec2d(gridPos)
	self.pathMatrix.setByXY(wpGrid.x, wpGrid.y, 0)
	while ( (wpGrid.x+navX >= 0) and (wpGrid.x+navX < GRIDCOLS) and (wpGrid.y+navY >=0) and (wpGrid.y+navY < GRIDROWS) 
		and self.pathMatrix.getByXY(wpGrid.x+navX, wpGrid.y+navY) == 1):
	    wpGrid = vec2d(wpGrid.x+navX, wpGrid.y+navY)
	    self.pathMatrix.setByXY(wpGrid.x, wpGrid.y, 0)

	self.waypoint = vec2d( gridToScreen( (wpGrid.x, wpGrid.y) ) )
	self.waypoint += GRIDSIZE/2
	return False
	
	
	
    def update(self, time_passed):
        self.image = pygame.transform.rotate(self.base_images[self.animFrame], -self.direction.angle)

	self._animate(time_passed)    # maybe time to go to next animation frame
	
        # Compute and apply the displacement to the position 
        # vector. The displacement is a vector, having the angle
        # of self.direction (which is normalized to not affect
        # the magnitude of the displacement)
        #
        displacement = vec2d(    
            self.direction.x * self.speed * time_passed,
            self.direction.y * self.speed * time_passed)
        
	oldPos = vec2d(self.pos.x, self.pos.y)
        self.pos += displacement
	
	# # # IF GONE TOO FAR # # 
	#print displacement
	#print "%s  --  %s" %(oldPos, self.pos)
	#print (abs(oldPos.x - self.waypoint.x)+abs(oldPos.y - self.waypoint.y))
	#print (abs(self.pos.x - self.waypoint.x)+abs(self.pos.y - self.waypoint.y))
	#print "--"
	bGoalReached = False
	if ( abs(oldPos.x - self.waypoint.x)+abs(oldPos.y - self.waypoint.y) < 
	     abs(self.pos.x - self.waypoint.x)+abs(self.pos.y - self.waypoint.y) ):
	    self.pos = self.waypoint
	    #print "navigating"
	    bGoalReached = self.navigate()
	
	self.image_w, self.image_h = self.image.get_size()
	return bGoalReached
	
    def blitme(self):
	# The creep image is placed at self.pos.
	# To allow for smooth movement even when the creep rotates
	# and the image size changes, its placement is always
	# centered.
	#
	draw_pos = self.image.get_rect().move(
	    self.pos.x - self.image_w / 2, 
	    self.pos.y - self.image_h / 2)
	self.screen.blit(self.image, draw_pos)
	
    #------------------ PRIVATE PARTS ------------------#
    _animCounter = 0
    def _animate(self, time_passed):
        self._animCounter += time_passed
        if self._animCounter > 60:
	    self._animCounter = 0
	    self.animFrame += 1
	    if self.animFrame >= len(self.base_images):
		self.animFrame = 0
	    
				     
	
class Enumerate(object):
    def __init__(self, names):
	for number, name in enumerate(names.split()):
	    setattr(self, name, number)  

class Matrix(object):
    def __init__(self, cols, rows):
	self.cols = cols
	self.rows = rows

	self.matrix = []
	for x in range(rows):
	    ea_row = []
	    for y in range(cols):
		ea_row.append(0)
	    self.matrix.append(ea_row)
    def deepCopy(self):
	copy = Matrix(self.cols, self.rows)
	for y in range (self.rows):
	    for x in range (self.cols):
		copy.setByXY(x, y, self.getByXY(x, y))
	return copy
  
    def setItem(self, col, row, v):
	self.matrix[col][row] = v
    def getItem(self, col, row):
	return self.matrix[col][row]
    def setByXY(self, x, y, v):
	self.matrix[y][x] = v
    def getByXY(self, x, y):
	return self.matrix[y][x]
  
    def __repr__(self):
	outStr = ""
	for i in range(self.rows):
	    outStr += 'Row %s = %s\n' % (i, self.matrix[i])
	return outStr
    
    def printMaze(self):
	outStr = ""
	for i in range(self.rows):
	    for j in range(self.cols):
		outStr += '%s ' % (self.getItem(i,j))
	    outStr += '\n'
	print outStr
    


class Grid(object):

    def __init__(self, cols, rows):
	self.grid = Matrix(cols, rows)
    
    def addTower(self, (clickX, clickY)):
	# print (clickX, clickY)
	(gridX, gridY) = screenToGrid((clickX, clickY))
	# print (gridX, gridY)
	if ((gridX > 0) and (gridX < self.grid.cols) and (gridY > 0) and (gridY < self.grid.rows) and 
	    ((self.grid.getItem(gridY, gridX)  + self.grid.getItem(gridY-1, gridX) + self.grid.getItem(gridY, gridX-1) + self.grid.getItem(gridY-1, gridX-1)) == 0)):
	    # 2 is upper left corner of tower
	    # 1 is any other part of the tower
	    self.grid.setItem(gridY, gridX, 1)
	    self.grid.setItem(gridY-1, gridX, 1)
	    self.grid.setItem(gridY, gridX-1, 1)
	    self.grid.setItem(gridY-1, gridX-1, 2)    
    def removeTower(self, (clickX, clickY)):
	# print (clickX, clickY)
	(gridX, gridY) = screenToGrid((clickX, clickY))
	# print (gridX, gridY)
	if ((gridX > 0) and (gridX < self.grid.cols) and (gridY > 0) and (gridY < self.grid.rows)):
	    self.grid.setItem(gridY, gridX, 0)
	    self.grid.setItem(gridY-1, gridX, 0)
	    self.grid.setItem(gridY, gridX-1, 0)
	    self.grid.setItem(gridY-1, gridX-1, 0)
	    
	    
	    
    def mz_getNodeContents(self, iNodeNo, matrix=None):
	if (matrix == None):
	    matrix = self.grid
	iCols = matrix.cols
	return matrix.getItem(iNodeNo/iCols, iNodeNo-iNodeNo/iCols*iCols)
    def mz_changeNodeContents(self, iNodeNo, iNewValue, matrix=None):
	if (matrix == None):
	    matrix = self.grid
	iCols = matrix.cols
	matrix.setItem(iNodeNo/iCols, iNodeNo-iNodeNo/iCols*iCols, iNewValue)
    def mz_findPath(self, iFromY, iFromX, iToY, iToX):
	iBeginningNode = iFromY*self.grid.cols + iFromX
	iEndingNode = iToY*self.grid.cols + iToX
	return (self.mz_search(iBeginningNode, iEndingNode))
    def mz_search(self, iStart, iStop):
	Status = Enumerate('Ready Waiting Processed')
	empty = 0
	iRows = self.grid.rows
	iCols = self.grid.cols
	iMax = iRows*iCols
	Queue = [None]*iMax
	Origin = [None]*iMax
	iFront = 0
	iRear = 0
	# check if starting and end points are valid
	if ( self.mz_getNodeContents(iStart) != empty ) or ( self.mz_getNodeContents(iStop) != empty):
	    print "start or end point invalid"
	    return
	
	iMazeStatus = Matrix(iCols, iRows)
	for i in range (0, iRows-1):
	    for j in range (0, iCols-1):
		iMazeStatus.setItem(i, j, Status.Ready)
		
	Queue[iRear] = iStart
	Origin[iRear] = -1
	iRear += 1
	
	iCurrent = iLeft = iRight = iTop = iDown = iRightUp = iRightDown = iLeftUp = iLeftDown = 0
	while (iFront != iRear):
	    if (Queue[iFront]==iStop):    # maze is solved
		break
	    
	    iCurrent = Queue[iFront]
	    
	    iLeft = iCurrent - 1    # # # # # CHECK LEFT # # # # #
	    if (iLeft>0 and iLeft/iCols==iCurrent/iCols):    #if left node exists
		if (self.mz_getNodeContents(iLeft) == empty):    #if left node is clear for path
		    if (self.mz_getNodeContents(iLeft, iMazeStatus) == Status.Ready):
			Queue[iRear] = iLeft
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iLeft, Status.Waiting, iMazeStatus)    #change status to waiting
			iRear+=1
	    iRight = iCurrent + 1    # # # # # CHECK RIGHT # # # # #
	    if (iRight<iMax and iRight/iCols==iCurrent/iCols):    
		if (self.mz_getNodeContents(iRight) == empty): 
		    if (self.mz_getNodeContents(iRight, iMazeStatus) == Status.Ready):
			Queue[iRear] = iRight
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iRight, Status.Waiting, iMazeStatus)
			iRear+=1
	    iTop = iCurrent - iCols    # # # # # CHECK UP # # # # #
	    if (iTop >= 0):    # if top node exists
		if (self.mz_getNodeContents(iTop) == empty):    #if top node is clear for path
		    if (self.mz_getNodeContents(iTop, iMazeStatus) == Status.Ready):    #if top node is ready
			Queue[iRear] = iTop    #add to queue
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iTop, Status.Waiting, iMazeStatus)
			iRear+=1
	    iDown = iCurrent + iCols    # # # # # CHECK DOWN # # # # #
	    if (iDown < iMax):    # if bottom node exists
		if (self.mz_getNodeContents(iDown) == empty):    #if bottom node is clear for path
		    if (self.mz_getNodeContents(iDown, iMazeStatus) == Status.Ready):    #if bottom node is ready
			Queue[iRear] = iDown
			Origin[iRear] = iCurrent
			self. mz_changeNodeContents(iDown, Status.Waiting, iMazeStatus)
			iRear+=1
	    iRightDown = iCurrent + iCols + 1    # # # # # DOWN RIGHT # # # # #
	    if (iRightDown<iMax and iRightDown>=0 and iRightDown/iCols==iCurrent/iCols+1):    #if bottom right node exists
		if (self.mz_getNodeContents(iRightDown)==empty and    #if is clear for path
		    self.mz_getNodeContents(iRightDown-1)==empty and    #and adjacent corners...
		    self.mz_getNodeContents(iRightDown-iCols)==empty):    
		    if (self.mz_getNodeContents(iRightDown, iMazeStatus) == Status.Ready):    #if ready
			Queue[iRear] = iRightDown
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iRightDown, Status.Waiting, iMazeStatus)
			iRear+=1
	    iRightUp = iCurrent - iCols + 1    # # # # # UP RIGHT # # # # #
	    if (iRightUp>=0 and iRightUp<iMax and iRightUp/iCols==iCurrent/iCols-1):
		if (self.mz_getNodeContents(iRightUp) == empty and
		    self.mz_getNodeContents(iRightUp-1) == empty and
		    self.mz_getNodeContents(iRightUp+iCols) == empty):
		    if (self.mz_getNodeContents(iRightUp, iMazeStatus) == Status.Ready):
			Queue[iRear] = iRightUp
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iRightUp, Status.Waiting,iMazeStatus)
			iRear+=1
	    iLeftDown = iCurrent + iCols -1    # # # # # DOWN LEFT # # # # #
	    if (iLeftDown<iMax and iLeftDown>=0 and iLeftDown/iCols==iCurrent/iCols+1):
		if (self.mz_getNodeContents(iLeftDown) == empty and
		    self.mz_getNodeContents(iLeftDown+1) == empty and
		    self.mz_getNodeContents(iLeftDown-iCols) == empty):
		    if (self.mz_getNodeContents(iLeftDown, iMazeStatus) == Status.Ready):
			Queue[iRear] = iRightDown
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iLeftDown, Status.Waiting, iMazeStatus)
			iRear += 1
	    iLeftUp = iCurrent - iCols - 1    # # # # # UP LEFT # # # # #
	    if (iLeftUp>=0 and iLeftUp<iMax and iLeftUp/iCols==iCurrent/iCols-1):
		if (self.mz_getNodeContents(iLeftUp) == empty and
		    self.mz_getNodeContents(iLeftUp+1) == empty and
		    self.mz_getNodeContents(iLeftUp+iCols) == empty):
		    if (self.mz_getNodeContents(iLeftUp, iMazeStatus) == Status.Ready):
			Queue[iRear] = iLeftUp
			Origin[iRear] = iCurrent
			self.mz_changeNodeContents(iLeftUp, Status.Waiting, iMazeStatus)
			iRear+=1
			
	    self.mz_changeNodeContents(iCurrent, Status.Processed, iMazeStatus)
	    iFront+=1
	
	# create an array for the solution
	iMazeSolved = Matrix(self.grid.cols, self.grid.rows)
	# make a path in the solved maze
	iCurrent = iStop
	self.mz_changeNodeContents(iCurrent, 1, iMazeSolved)
	for i in range (iFront, -1, -1):
	    if (Queue[i] == iCurrent):
		iCurrent = Origin[i]
		if (iCurrent == -1):    #maze is solved
		    return iMazeSolved		
		self.mz_changeNodeContents(iCurrent, 1, iMazeSolved)
	return None    #no path exists
	    
	
def run_game():
    pygame.init()
    clock = pygame.time.Clock()

    screen = pygame.display.set_mode(SCREEN_DIMENSION)
    background = pygame.image.load(BGROUND_FILENAME).convert()
    backgroundRect = background.get_rect()

    wallImg = pygame.image.load('wall40.jpg').convert()
    greenImg = pygame.image.load('green40.png').convert_alpha()
    
    gameGrid = Grid(GRIDCOLS, GRIDROWS)
    currentAnts = []
    currentPathToFinish = gameGrid.mz_findPath(0, 0, GRIDROWS-1, GRIDCOLS-1)

    # Main Game Loop
    while True:
        # Limit frame speed to 30 FPS
        time_passed = clock.tick(30)
        
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                exit_game()
	    if event.type == MOUSEBUTTONDOWN:
		if event.button == 1:    # left-click = add a tower
		    gameGrid.addTower(event.pos)
		    oldPathToFinish = currentPathToFinish
		    currentPathToFinish=gameGrid.mz_findPath(0, 0, GRIDROWS-1, GRIDCOLS-1)
		    if currentPathToFinish==None:
			gameGrid.removeTower(event.pos)
			currentPathToFinish = oldPathToFinish
		elif event.button == 3:    # right-click = add an ant
		    currentAnts.append(Attacker(screen, ANT_FILENAMES, currentPathToFinish, 0.15, 100))
		    
        # Redraw the background
	screen.blit(background, backgroundRect)

	# Draw a green square where tower will go
	(mouseX, mouseY) = pygame.mouse.get_pos()
	if ((mouseX > GRIDSTART_X+GRIDSIZE) and (mouseY > GRIDSTART_Y+GRIDSIZE) and (mouseY < GRIDSTART_Y+GRIDSIZE*GRIDROWS) and (mouseX < GRIDSTART_X+GRIDSIZE*GRIDCOLS)):
	    tempx = (mouseX-GRIDSTART_X+GRIDSIZE)/GRIDSIZE*GRIDSIZE - 15
	    tempy = (mouseY-GRIDSTART_Y+GRIDSIZE)/GRIDSIZE*GRIDSIZE - 15
	    screen.blit(greenImg, (tempx, tempy))

	# Loop over gameGrid, drawing walls on 2s
	for i in range(gameGrid.grid.rows):
	    for j in range(gameGrid.grid.cols):
		if gameGrid.grid.getItem(i, j) == 2:
		    screen.blit(wallImg, gridToScreen((j, i)))
	
	# update ants
	for ant in currentAnts:
	    bGoalReached = ant.update(time_passed)
	    ant.blitme()
	    if bGoalReached:
		currentAnts.remove(ant)

        pygame.display.flip()
	

def exit_game():
    sys.exit()

run_game()

